@@ -1,39 +1,40 @@
-import re
-import json
-import time
-import socket
-import base64
-
 from .common import (
     Chat,
     BaseChatDownloader,
-    Timeout,
-    Remapper as r
+    Remapper as r,
+    Image
 )
 
-from requests.exceptions import RequestException
-
 from ..errors import (
     TwitchError,
-    UnexpectedHTML,
     NoChatReplay,
-    TimeoutException,
-    VideoUnavailable
+    VideoUnavailable,
+    UserNotFound
 )
 
 from ..utils import (
     ensure_seconds,
     timestamp_to_microseconds,
     seconds_to_time,
-    try_get,
     int_or_none,
     replace_with_underscores,
     multi_get,
-    log,
     remove_prefixes,
     attempts
 )
 
+from ..debugging import log
+
+import re
+import json
+import time
+import socket
+import base64
+
+from requests.exceptions import RequestException
+from json.decoder import JSONDecodeError
+
+
 # TODO export as another module?
 
 
@@ -104,6 +105,7 @@ class TwitchChatDownloader(BaseChatDownloader):
         super().__init__(**kwargs)
 
         # Only get badge info if not already retrieved
+        # TODO add argument (no_badges)
         if not TwitchChatDownloader._BADGE_INFO:
             TwitchChatDownloader._BADGE_INFO = self._session_get_json(
                 self._BADGE_INFO_URL).get('badge_sets') or {}
@@ -115,9 +117,6 @@ class TwitchChatDownloader(BaseChatDownloader):
             'params': {
                 'url': 'https://www.twitch.tv/xqcow',
                 'timeout': 5
-            },
-            'expected_result': {
-                'error': TimeoutException
             }
         },
 
@@ -182,30 +181,29 @@ class TwitchChatDownloader(BaseChatDownloader):
         'format': 'twitch',
     }
 
-    _VALID_URL = r'https?://(?:(?:www|go|m|clips)\.)?twitch\.tv'
-
-    # e.g. 'http://www.twitch.tv/riotgames/v/6528877?t=5m10s'
-    _VALID_VOD_URL = r'''(?x)
+    _VALID_URLS = {
+        # e.g. 'http://www.twitch.tv/riotgames/v/6528877?t=5m10s'
+        '_get_chat_by_vod_id': r'''(?x)
                     https?://
                         (?:
                             (?:(?:www|go|m)\.)?twitch\.tv/(?:[^/]+/v(?:ideo)?|videos)/|
                             player\.twitch\.tv/\?.*?\bvideo=v?
                         )
                         (?P<id>\d+)
-                    '''
+                    ''',
 
-    # e.g. 'https://clips.twitch.tv/FaintLightGullWholeWheat'
-    _VALID_CLIPS_URL = r'''(?x)
+        # e.g. 'https://clips.twitch.tv/FaintLightGullWholeWheat'
+        '_get_chat_by_clip_id': r'''(?x)
                         https?://
                             (?:
                                 clips\.twitch\.tv/(?:embed\?.*?\bclip=|(?:[^/]+/)*)|
                                 (?:(?:www|go|m)\.)?twitch\.tv/[^/]+/clip/
                             )
                             (?P<id>[^/?#&]+)
-                        '''
+                        ''',
 
-    # e.g. 'http://www.twitch.tv/shroomztv'
-    _VALID_STREAM_URL = r'''(?x)
+        # e.g. 'http://www.twitch.tv/shroomztv'
+        '_get_chat_by_stream_id': r'''(?x)
                         https?://
                             (?:
                                 (?:(?:www|go|m)\.)?twitch\.tv/|
@@ -213,6 +211,7 @@ class TwitchChatDownloader(BaseChatDownloader):
                             )
                             (?P<id>[^/#?]+)
                         '''
+    }
 
     _CLIENT_ID = 'kimne78kx3ncx6brgo4mv6wki5h1ko'  # public client id
 
@@ -230,31 +229,32 @@ class TwitchChatDownloader(BaseChatDownloader):
     }
 
     @staticmethod
-    def parse_bool(text):
+    def _parse_bool(text):
         return text == '1'
 
     @staticmethod
-    def parse_bool_text(text):
+    def _parse_bool_text(text):
         return text == 'true'
 
     @staticmethod
-    def parse_author_images(original_url):
+    def _parse_author_images(original_url):
+        # e.g. https://static-cdn.jtvnw.net/jtv_user_pictures/3892c956-0616-4fc9-b2fe-527b1be0b623-profile_image-300x300.png
         smaller_icon = original_url.replace('300x300', '70x70')
         return [
-            BaseChatDownloader.create_image(original_url, 300, 300),
-            BaseChatDownloader.create_image(smaller_icon, 70, 70),
+            Image(original_url, 300, 300).json(),
+            Image(smaller_icon, 70, 70).json(),
         ]
 
     @staticmethod
-    def parse_commenter(commenter):
+    def _parse_commenter(commenter):
         info = {}
         for key in commenter or []:
-            BaseChatDownloader.remap(
-                info, TwitchChatDownloader._AUTHOR_REMAPPING, key, commenter[key])
+            r.remap(info, TwitchChatDownloader._AUTHOR_REMAPPING,
+                    key, commenter[key])
         return info
 
     @staticmethod
-    def parse_message_info(message):
+    def _parse_message_info(message):
         # Ignore: fragments, is_action
         message_text = message.get('body') or ''
 
@@ -270,8 +270,8 @@ class TwitchChatDownloader(BaseChatDownloader):
                 if emote_id not in message_emotes:
                     message_emotes[emote_id] = {
                         'id': emote_id,
-                        'images': TwitchChatDownloader.generate_twitch_emote_image_list(emote_id),
-                        'name': message_text[begin:end+1]
+                        'images': TwitchChatDownloader._generate_emote_image_list(emote_id),
+                        'name': message_text[begin:end + 1]
                     }
 
                     locations[emote_id] = []
@@ -298,24 +298,24 @@ class TwitchChatDownloader(BaseChatDownloader):
         return message_info
 
     @staticmethod
-    def decode_pseudo_BNF(text):
+    def _decode_pseudo_BNF(text):
         """
         Decode text according to https://ircv3.net/specs/extensions/message-tags.html
         """
         return text.replace(r'\:', ';').replace(r'\s', ' ')
 
     @staticmethod
-    def generate_twitch_emote_image_list(emote_id):
+    def _generate_emote_image_list(emote_id):
         emote_image_list = []
         for theme in ('light', 'dark'):
             for size in ((28, '1.0'), (56, '2.0'), (112, '3.0')):
-                image = BaseChatDownloader.create_image(
+                image = Image(
                     TwitchChatDownloader._EMOTE_URL_TEMPLATE.format(
                         emote_id, theme, size[1]),
                     size[0],
                     size[0],
                     '{0}x{0}-{1}'.format(size[0], theme)
-                )
+                ).json()
 
                 emote_image_list.append(image)
         return emote_image_list
@@ -324,7 +324,7 @@ class TwitchChatDownloader(BaseChatDownloader):
     _EMOTE_URL_TEMPLATE = 'https://static-cdn.jtvnw.net/emoticons/v2/{}/default/{}/{}'
 
     @staticmethod
-    def parse_emotes(text):
+    def _parse_emotes(text):
         # Information to replace text in the message with emote images. This can be empty.
         # <emote ID>:<first index>-<last index>,<another first index>-<another last index>/<another emote ID>:<first index>-<last index>
         emotes = []
@@ -336,7 +336,7 @@ class TwitchChatDownloader(BaseChatDownloader):
             emote = {
                 'id': emote_id,
                 'locations': match[1].split(','),
-                'images': TwitchChatDownloader.generate_twitch_emote_image_list(emote_id)
+                'images': TwitchChatDownloader._generate_emote_image_list(emote_id)
             }
             emotes.append(emote)
 
@@ -346,7 +346,7 @@ class TwitchChatDownloader(BaseChatDownloader):
         '_id': r('id', int_or_none),
         'name': 'name',
         'display_name': 'display_name',
-        'logo': r('images', parse_author_images),
+        'logo': r('images', _parse_author_images),
         'type': 'type',
         'created_at': r('created_at', timestamp_to_microseconds),
         # 'updated_at': r('updated_at', 'parse_timestamp'),
@@ -356,14 +356,14 @@ class TwitchChatDownloader(BaseChatDownloader):
     _COMMENT_REMAPPING = {
         '_id': 'message_id',
         'created_at': r('timestamp', timestamp_to_microseconds),
-        'commenter': r('author', parse_commenter),
+        'commenter': r('author', _parse_commenter),
 
         'content_offset_seconds': 'time_in_seconds',
 
         'source': 'source',
         'state': 'state',
         # TODO make sure body vs. fragments okay
-        'message': r(None, parse_message_info, True)
+        'message': r(None, _parse_message_info, True)
     }
 
     _MESSAGE_PARAM_REMAPPING = {
@@ -387,10 +387,11 @@ class TwitchChatDownloader(BaseChatDownloader):
         'msg-param-sender-login': 'gifter_name',
         'msg-param-sender-name': 'gifter_display_name',
 
-        'msg-param-should-share-streak': r('user_wants_to_share_streaks', parse_bool),
+        'msg-param-should-share-streak': r('user_wants_to_share_streaks', _parse_bool),
         'msg-param-streak-months': r('number_of_consecutive_months_subscribed', int_or_none),
         'msg-param-sub-plan': r('subscription_type', lambda x: TwitchChatDownloader._SUBSCRIPTION_TYPES.get(x)),
-        'msg-param-sub-plan-name': r('subscription_plan_name', decode_pseudo_BNF),
+        'msg-param-sub-plan-name': r('subscription_plan_name', _decode_pseudo_BNF),
+        'msg-param-sub-benefit-end-month': r('sub_benefit_end_month', int_or_none),
 
         'msg-param-ritual-name': 'ritual_name',
 
@@ -402,12 +403,12 @@ class TwitchChatDownloader(BaseChatDownloader):
         # resub
         'msg-param-multimonth-duration': r('multimonth_duration', int_or_none),
         'msg-param-multimonth-tenure': r('multimonth_tenure', int_or_none),
-        'msg-param-was-gifted': r('was_gifted', parse_bool_text),
+        'msg-param-was-gifted': r('was_gifted', _parse_bool_text),
 
         'msg-param-gifter-id': 'gifter_id',
         'msg-param-gifter-login': 'gifter_name',
         'msg-param-gifter-name': 'gifter_display_name',
-        'msg-param-anon-gift': r('was_anonymous_gift', parse_bool_text),
+        'msg-param-anon-gift': r('was_anonymous_gift', _parse_bool_text),
         'msg-param-gift-month-being-redeemed': r('gift_months_being_redeemed', int_or_none),
 
         # rewardgift
@@ -418,12 +419,12 @@ class TwitchChatDownloader(BaseChatDownloader):
         'msg-param-trigger-amount': r('trigger_amount', int_or_none),
 
         # submysterygift
-        'msg-param-origin-id': r('origin_id', decode_pseudo_BNF),
+        'msg-param-origin-id': r('origin_id', _decode_pseudo_BNF),
         'msg-param-sender-count': r('sender_count', int_or_none),
         'msg-param-mass-gift-count': r('mass_gift_count', int_or_none),
 
         # communitypayforward
-        'msg-param-prior-gifter-anonymous': r('prior_gifter_anonymous', parse_bool_text),
+        'msg-param-prior-gifter-anonymous': r('prior_gifter_anonymous', _parse_bool_text),
         'msg-param-prior-gifter-user-name': 'prior_gifter_name',
         'msg-param-prior-gifter-display-name': 'prior_gifter_display_name',
         'msg-param-prior-gifter-id': 'prior_gifter_id',
@@ -500,13 +501,13 @@ class TwitchChatDownloader(BaseChatDownloader):
         'bits': r('bits', int_or_none),
 
         'id': 'message_id',
-        'mod': r('author_is_moderator', parse_bool),
+        'mod': r('author_is_moderator', _parse_bool),
         'room-id': r('channel_id', int_or_none),
 
         'tmi-sent-ts': r('timestamp', lambda x: int_or_none(x, 0) * 1000),
 
-        'subscriber': r('author_is_subscriber', parse_bool),
-        'turbo': r('author_is_turbo', parse_bool),
+        'subscriber': r('author_is_subscriber', _parse_bool),
+        'turbo': r('author_is_turbo', _parse_bool),
 
         'client-nonce': 'client_nonce',
 
@@ -514,7 +515,7 @@ class TwitchChatDownloader(BaseChatDownloader):
 
 
 
-        'reply-parent-msg-body': r('in_reply_to_message', decode_pseudo_BNF),
+        'reply-parent-msg-body': r('in_reply_to_message', _decode_pseudo_BNF),
         'reply-parent-user-id': r('in_reply_to_author_id', int_or_none),
         'reply-parent-msg-id': 'in_reply_to_message_id',
         'reply-parent-display-name': 'in_reply_to_author_display_name',
@@ -524,22 +525,22 @@ class TwitchChatDownloader(BaseChatDownloader):
         'custom-reward-id': 'custom_reward_id',
 
 
-        'emotes': r('emotes', parse_emotes),
+        'emotes': r('emotes', _parse_emotes),
         'flags': 'flags',
 
 
 
         # ROOMSTATE
-        'emote-only': r('emote_only', parse_bool),
+        'emote-only': r('emote_only', _parse_bool),
         'followers-only': r('follower_only', int_or_none),
 
-        'r9k': r('r9k_mode', parse_bool),
+        'r9k': r('r9k_mode', _parse_bool),
         'slow': r('slow_mode', int_or_none),
-        'subs-only': r('subscriber_only', parse_bool),
-        'rituals': r('rituals_enabled', parse_bool),
+        'subs-only': r('subscriber_only', _parse_bool),
+        'rituals': r('rituals_enabled', _parse_bool),
 
         # USERNOTICE
-        'system-msg': r('system_message', decode_pseudo_BNF),
+        'system-msg': r('system_message', _decode_pseudo_BNF),
 
         # (Commands)
         # HOSTTARGET
@@ -772,13 +773,13 @@ class TwitchChatDownloader(BaseChatDownloader):
     }
 
     _MESSAGE_TYPE_REMAPPING = {}
-    for message_group in _MESSAGE_GROUP_REMAPPINGS:
-        value = _MESSAGE_GROUP_REMAPPINGS[message_group]
-        _MESSAGE_TYPE_REMAPPING.update(value)
+    for _message_group in _MESSAGE_GROUP_REMAPPINGS:
+        _value = _MESSAGE_GROUP_REMAPPINGS[_message_group]
+        _MESSAGE_TYPE_REMAPPING.update(_value)
 
-        if message_group not in _MESSAGE_GROUPS:
-            _MESSAGE_GROUPS[message_group] = []
-        _MESSAGE_GROUPS[message_group] += list(value.values())
+        if _message_group not in _MESSAGE_GROUPS:
+            _MESSAGE_GROUPS[_message_group] = []
+        _MESSAGE_GROUPS[_message_group] += list(_value.values())
 
     _SUBSCRIBER_BADGE_INFO = {}  # local cache for subscriber badge info
     _SUBSCRIBER_BADGE_URL = 'https://badges.twitch.tv/v1/badges/channels/{}/display'
@@ -800,8 +801,8 @@ class TwitchChatDownloader(BaseChatDownloader):
     def _parse_item(item, offset):
         info = {}
         for key in item:
-            BaseChatDownloader.remap(
-                info, TwitchChatDownloader._COMMENT_REMAPPING, key, item[key])  # , True
+            r.remap(info, TwitchChatDownloader._COMMENT_REMAPPING,
+                    key, item[key])  # , True
 
         if 'time_in_seconds' in info:
             info['time_in_seconds'] -= offset
@@ -813,21 +814,21 @@ class TwitchChatDownloader(BaseChatDownloader):
 
         badges = info.pop('author_badges', None)
         if badges:
-            info['author']['badges'] = list(map(lambda x: TwitchChatDownloader.parse_badge_info(
+            info['author']['badges'] = list(map(lambda x: TwitchChatDownloader._parse_badge_info(
                 x.get('_id'), x.get('version'), channel_id), badges))
 
         user_notice_params = info.pop('user_notice_params', {})
 
         for key in user_notice_params:
-            BaseChatDownloader.remap(
-                info, TwitchChatDownloader._MESSAGE_PARAM_REMAPPING, key, user_notice_params[key], True)
+            r.remap(info, TwitchChatDownloader._MESSAGE_PARAM_REMAPPING,
+                    key, user_notice_params[key], True)
 
         # TODO add user colour to author dict
         # TODO check this works
         # author_colour
         # print()
 
-        BaseChatDownloader.move_to_dict(info, 'author')
+        BaseChatDownloader._move_to_dict(info, 'author')
 
         original_message_type = info.get('message_type')
         if original_message_type:
@@ -846,10 +847,10 @@ class TwitchChatDownloader(BaseChatDownloader):
         'ChannelVideoShelvesQuery': 'fb663273aa958ebe2f58d5fcb3aacc112d67ebfd7f414b095c5d1498d21aad92',
         'ClipsCards__User': 'b73ad2bfaecfd30a9e6c28fada15bd97032c83ec77a0440766a56fe0bd632777',
         'VideoMetadata': '226edb3e692509f727fd56821f5653c05740242c82b0388883e0c0e75dcbf687',
+        'FilterableVideoTower_Videos': 'a937f1d22e269e39a03b509f65a7490f9fc247d7f83d6ac1421523e3b68042cb',
 
         # Not used yet
         # 'CollectionSideBar': '27111f1b382effad0b6def325caef1909c733fe6a4fbabf54f8d491ef2cf2f14',
-        # 'FilterableVideoTower_Videos': 'a937f1d22e269e39a03b509f65a7490f9fc247d7f83d6ac1421523e3b68042cb',
         # 'ChannelCollectionsContent': '07e3691a1bad77a36aba590c351180439a40baefc1c275356f40fc7082419a84',
         # 'ComscoreStreamingQuery': 'e1edae8122517d013405f237ffcc124515dc6ded82480a88daef69c83b53ac01',
         # 'VideoPreviewOverlay': '3006e77e51b128d838fa4e835723ca4dc9a05c5efd4466c1085215c6e437e65c',
@@ -871,8 +872,54 @@ class TwitchChatDownloader(BaseChatDownloader):
             }
         return self._download_base_gql(ops)
 
-    def get_user_clips(self, user_name, limit=100, filter_by='LAST_WEEK'):
+    _USER_REMAPPING = {
+        'id': 'id',
+        'login': 'name',
+        'displayName': 'display_name',
+        'profileImageURL': 'profile_image_url',
+        'primaryColorHex': 'colour'
+    }
 
+    @staticmethod
+    def _parse_user(item):
+        if isinstance(item, dict):
+            return r.remap_dict(item, TwitchChatDownloader._USER_REMAPPING)
+        return None
+
+    _GAME_REMAPPING = {
+        'id': 'id',
+        'name': 'name',
+        'displayName': 'display_name',
+        'boxArtURL': 'box_art_url'
+    }
+
+    @staticmethod
+    def _parse_game(item):
+        if isinstance(item, dict):
+            return r.remap_dict(item, TwitchChatDownloader._GAME_REMAPPING)
+        return None
+
+    _CLIP_REMAPPING = {
+        'id': r('id', int_or_none),
+        'slug': 'slug',
+        'url': 'url',
+        'embedURL': 'embed_url',
+        'title': 'title',
+        'viewCount': 'views',
+        'language': 'language',
+        'curator': r('curator', _parse_user),
+        'game': r('game', _parse_game),
+        'language': 'language',
+        'broadcaster': r('broadcaster', _parse_user),
+
+        'thumbnailURL': 'thumbnail_url',
+        'createdAt': r('created_at', timestamp_to_microseconds),
+        'durationSeconds': 'duration',
+
+
+    }
+
+    def get_user_clips(self, username, limit=100, filter_by='LAST_WEEK'):
         # filter_by:
         # LAST_WEEK
         # LAST_DAY
@@ -881,25 +928,20 @@ class TwitchChatDownloader(BaseChatDownloader):
 
         # offset= 20# , offset = 0
 
-        count = limit
+        remaining_count = limit
         offset = 0
-        to_return = []
         while True:
-            cursor = base64.b64encode(str(offset).encode()).decode()
-            num_to_get = max(min(count, 100), 0)
+            num_to_get = max(min(remaining_count, 100), 0)  # in this call
             if num_to_get <= 0:
                 break
-            # print('num_to_get', num_to_get)
 
             query = [{
                 'operationName': 'ClipsCards__User',
-                'variables':
-                {
-                    'cursor': cursor,
-                    'login': user_name,
+                'variables': {
+                    'cursor': base64.b64encode(str(offset).encode()).decode(),
+                    'login': username,
                     'limit': num_to_get,
-                    'criteria':
-                    {
+                    'criteria': {
                         'filter': filter_by
                     }
                 }
@@ -907,29 +949,93 @@ class TwitchChatDownloader(BaseChatDownloader):
             info = self._download_gql(query)
             if not info:
                 break
+
             clips = info[0]['data']['user']['clips']
 
             edges = clips['edges']
+            remaining_count -= len(edges)
 
-            count -= len(edges)
-
-            to_return += edges
+            for edge in edges:
+                node = edge['node'] or {}
+                yield r.remap_dict(node, TwitchChatDownloader._CLIP_REMAPPING)
 
             if not clips['pageInfo']['hasNextPage']:
                 break
 
-        return to_return
+    _VIDEO_REMAPPING = {
+        'id': r('id', int_or_none),
+        'animatedPreviewURL': 'animated_preview_url',
+        'game': r('game', _parse_game),
 
-        # TODO check if pageInfo hasNextPage
+        'lengthSeconds': 'duration',
 
-        # print(len(edges))
-        return edges
+        'owner': r('owner', _parse_user),
 
-    def get_user_vods(self, user_name):
+        'previewThumbnailURL': 'preview_thumbnail_url',
+
+        'publishedAt': r('published_at', timestamp_to_microseconds),
+
+        'title': 'title',
+        'viewCount': 'views',
+        'resourceRestriction': 'resource_restriction'
+
+        # 'contentTags': 'tags'
+    }
+
+    def get_user_videos(self, username, limit=30, video_type=None, sort='TIME'):
+        # Name -> broadcastType
+
+        # Collections (ChannelCollectionsContent)
+        #
+
+        # Past broadcasts -> ARCHIVE
+        # Highlights -> HIGHLIGHT
+        # Uploads -> UPLOAD
+        # Past premieres -> PAST_PREMIERE
+        # All Videos -> null
+
+        # Sort -> videoSort
+        # VIEWS
+        # TIME
+
+        remaining_count = limit
+        # offset = 0
+        while True:
+            num_to_get = max(min(remaining_count, 30), 0)  # in this call
+            if num_to_get <= 0:
+                break
+
+            query = [{
+                'operationName': 'FilterableVideoTower_Videos',
+                'variables': {
+                    'limit': num_to_get,
+                    'channelOwnerLogin': username,
+                    'broadcastType': video_type,
+                    'videoSort': sort
+                }
+            }]
+
+            info = self._download_gql(query)
+            if not info:
+                break
+
+            videos = info[0]['data']['user']['videos']
+
+            edges = videos['edges']
+            remaining_count -= len(edges)
+
+            for edge in edges:
+                node = edge['node'] or {}
+                yield r.remap_dict(node, TwitchChatDownloader._VIDEO_REMAPPING)
+
+            if not videos['pageInfo']['hasNextPage']:
+                break
+
+    def get_featured_videos(self, username):
         query = [{
             'operationName': 'ChannelVideoShelvesQuery',
             'variables': {
-                'channelLogin': user_name,
+                'channelLogin': username,
                 'first': 5
             }
         }]
@@ -938,14 +1044,29 @@ class TwitchChatDownloader(BaseChatDownloader):
         # print(len(edges))
         return edges
 
+    _LIVESTREAM_REMAPPING = {
+
+        'id': r('id', int_or_none),
+        'title': 'title',
+        'viewersCount': 'viewers',
+
+        'previewImageURL': 'preview_image_url',
+        'broadcaster': r('broadcaster', _parse_user),
+        'game': r('game', _parse_game),
+
+
+        # 'tags': 'tags'
+
+        'type': 'type'  # e.g. 'live'
+    }
+
     def get_top_livestreams(self, limit=30):
-        count = limit
+        remaining_count = limit
 
         cursor = ''
 
-        to_return = []
         while True:
-            num_to_get = max(min(count, 30), 0)
+            num_to_get = max(min(remaining_count, 30), 0)
             if num_to_get <= 0:
                 break
             # print('num_to_get', num_to_get)
@@ -965,20 +1086,37 @@ class TwitchChatDownloader(BaseChatDownloader):
             edges = self._download_gql(query)[0]['data']['streams']['edges']
 
             cursor = edges[-1]['cursor']
-            count -= num_to_get
+            remaining_count -= num_to_get
 
-            to_return += edges
+            for edge in edges:
+                node = edge['node'] or {}
+                yield r.remap_dict(node, TwitchChatDownloader._LIVESTREAM_REMAPPING)
 
-        return to_return
-        # print(streams)
+    _TWITCH_HOME = 'https://www.twitch.tv'
+    _TWITCH_VIDEOS = 'https://www.twitch.tv/videos'
 
-    _REGEX_FUNCTION_MAP = [
-        (_VALID_VOD_URL, 'get_chat_by_vod_id'),
-        (_VALID_CLIPS_URL, 'get_chat_by_clip_id'),
-        (_VALID_STREAM_URL, 'get_chat_by_stream_id'),
-    ]
+    def generate_urls(self, livestream_limit=10, vod_limit=5, clip_limit=5, **kwargs):
+        # max_tests = livestream_limit + livestream_limit*(vod_limit+clip_limit)
+
+        livestreams = self.get_top_livestreams(livestream_limit)
+        for livestream in livestreams:
+            name = livestream['broadcaster']['name']
+
+            # e.g. https://www.twitch.tv/shroud
+            yield '{}/{}'.format(self._TWITCH_HOME, name)
+
+            vods = self.get_user_videos(name, vod_limit)
+            for vod in vods:
+                # e.g. https://www.twitch.tv/videos/12345678
+                yield '{}/{}'.format(self._TWITCH_VIDEOS, vod['id'])
+
+            clips = self.get_user_clips(name, clip_limit)
+            for clip in clips:
+                # e.g. https://clips.twitch.tv/FastThankfulLobsterEagleEye-SFi4SJWaTkAYu-B3
+                yield clip['url']
 
     # offset and max_duration are used by clips
+
     def _get_chat_messages_by_vod_id(self, vod_id, params, max_duration, offset=None):
 
         # twitch does not provide messages before the stream starts,
@@ -1006,17 +1144,9 @@ class TwitchChatDownloader(BaseChatDownloader):
         messages_groups_to_add = params.get('message_groups') or []
         messages_types_to_add = params.get('message_types') or []
 
-        def debug_log(*items):
-            log(
-                'debug',
-                items,
-                params.get('pause_on_debug')
-            )
-
         api_url = self._API_TEMPLATE.format(vod_id, self._CLIENT_ID)
 
         message_count = 0
-        timeout = Timeout(params.get('timeout'))
         # do not need inactivity timeout (not live)
         cursor = ''
         while True:
@@ -1024,11 +1154,10 @@ class TwitchChatDownloader(BaseChatDownloader):
                 api_url, cursor, content_offset_seconds)
 
             for attempt_number in attempts(max_attempts):
-                timeout.check_for_timeout()
                 try:
                     info = self._session_get_json(url)
                     break
-                except (UnexpectedHTML, RequestException) as e:
+                except (JSONDecodeError, RequestException) as e:
                     self.retry(attempt_number, max_attempts, e, retry_timeout)
 
             error_message = multi_get(info, 'error', 'message')
@@ -1044,13 +1173,14 @@ class TwitchChatDownloader(BaseChatDownloader):
                 missing_keys = data.keys() - TwitchChatDownloader._KNOWN_COMMENT_KEYS
 
                 if missing_keys:
-                    debug_log(
-                        'Missing keys found: {}'.format(missing_keys),
-                        'Original data: {}'.format(comment),
-                        'Parsed data: {}'.format(data),
-                        comment.keys(),
-                        TwitchChatDownloader._KNOWN_COMMENT_KEYS
-                    )
+                    self._debug_log(params,
+                                    'Missing keys found: {}'.format(
+                                        missing_keys),
+                                    'Original data: {}'.format(comment),
+                                    'Parsed data: {}'.format(data),
+                                    comment.keys(),
+                                    TwitchChatDownloader._KNOWN_COMMENT_KEYS
+                                    )
 
                 time_in_seconds = data.get('time_in_seconds', 0)
 
@@ -1062,7 +1192,7 @@ class TwitchChatDownloader(BaseChatDownloader):
                 elif after_end:  # after end
                     return  # while actually searching, if time is invalid
 
-                to_add = self.must_add_item(
+                to_add = self._must_add_item(
                     data,
                     self._MESSAGE_GROUPS,
                     messages_groups_to_add,
@@ -1082,6 +1212,9 @@ class TwitchChatDownloader(BaseChatDownloader):
             if not cursor:
                 return
 
+    def _get_chat_by_vod_id(self, match, params):
+        return self.get_chat_by_vod_id(match.group('id'), params)
+
     def get_chat_by_vod_id(self, vod_id, params):
         max_attempts = params.get('max_attempts')
         retry_timeout = params.get('retry_timeout')
@@ -1098,7 +1231,7 @@ class TwitchChatDownloader(BaseChatDownloader):
             try:
                 video = self._download_gql(query)[0]['data']['video']
                 break
-            except (UnexpectedHTML, RequestException) as e:
+            except (JSONDecodeError, RequestException) as e:
                 self.retry(attempt_number, max_attempts, e, retry_timeout)
 
         if not video:
@@ -1120,6 +1253,9 @@ class TwitchChatDownloader(BaseChatDownloader):
             is_live=False
         )
 
+    def _get_chat_by_clip_id(self, match, params):
+        return self.get_chat_by_clip_id(match.group('id'), params)
+
     def get_chat_by_clip_id(self, clip_id, params):
 
         max_attempts = params.get('max_attempts')
@@ -1133,7 +1269,7 @@ class TwitchChatDownloader(BaseChatDownloader):
             try:
                 clip = self._download_base_gql(query)['data']['clip']
                 break
-            except (UnexpectedHTML, RequestException) as e:
+            except (JSONDecodeError, RequestException) as e:
                 self.retry(attempt_number, max_attempts, e, retry_timeout)
 
         vod_id = multi_get(clip, 'video', 'id')
@@ -1174,7 +1310,7 @@ class TwitchChatDownloader(BaseChatDownloader):
     _BADGE_ID_REGEX = r'v1/([^/]+)/'
 
     @staticmethod
-    def parse_badge_info(name, version, channel_id):
+    def _parse_badge_info(name, version, channel_id):
         new_badge = {
             'name': replace_with_underscores(name),
             'version': int_or_none(version, version)
@@ -1182,10 +1318,9 @@ class TwitchChatDownloader(BaseChatDownloader):
 
         # prioritise custom emotes (e.g. subscriber and bits)
         channel_id = int(channel_id)
-        new_badge_info = try_get(TwitchChatDownloader._SUBSCRIBER_BADGE_INFO,
-                                 lambda x: x[channel_id][name]['versions'][version]
-                                 ) or try_get(
-                                     TwitchChatDownloader._BADGE_INFO, lambda x: x[name]['versions'][version])
+        new_badge_info = multi_get(
+            TwitchChatDownloader._SUBSCRIBER_BADGE_INFO, channel_id, name, 'versions', version) or multi_get(
+            TwitchChatDownloader._BADGE_INFO, name, 'versions', version)
 
         if new_badge_info:
             for key in TwitchChatDownloader._BADGE_KEYS:
@@ -1197,8 +1332,7 @@ class TwitchChatDownloader(BaseChatDownloader):
                 new_badge['icons'] = []
 
             for image_url, size in image_urls:
-                new_badge['icons'].append(
-                    BaseChatDownloader.create_image(image_url, size, size))
+                new_badge['icons'].append(Image(image_url, size, size).json())
 
             if image_urls:
                 badge_id = re.search(
@@ -1209,7 +1343,7 @@ class TwitchChatDownloader(BaseChatDownloader):
         return new_badge
 
     @staticmethod
-    def parse_irc_badges(badges, channel_id):
+    def _parse_irc_badges(badges, channel_id):
         info = []
         if not badges:
             return info
@@ -1229,7 +1363,7 @@ class TwitchChatDownloader(BaseChatDownloader):
                 ])
                 continue  # TODO debug
 
-            info.append(TwitchChatDownloader.parse_badge_info(
+            info.append(TwitchChatDownloader._parse_badge_info(
                 split[0], split[1], channel_id))
         return info
 
@@ -1255,7 +1389,7 @@ class TwitchChatDownloader(BaseChatDownloader):
             try:
                 first_location = list(
                     map(lambda x: int(x), emote['locations'][0].split('-')))
-                emote['name'] = message[first_location[0]:first_location[1]+1]
+                emote['name'] = message[first_location[0]:first_location[1] + 1]
             except Exception:
                 log('debug', [
                     'Invalid emote: {}'.format(emote),
@@ -1284,9 +1418,8 @@ class TwitchChatDownloader(BaseChatDownloader):
                 ])
                 continue
 
-            BaseChatDownloader.remap(info, TwitchChatDownloader._IRC_REMAPPING, keys[0], keys[1],
-                                     keep_unknown_keys=True,
-                                     replace_char_with_underscores='-')
+            r.remap(info, TwitchChatDownloader._IRC_REMAPPING,
+                    keys[0], keys[1], keep_unknown_keys=True, replace_char_with_underscores='-')
 
         message_match = match.group(3)
         if message_match:
@@ -1301,10 +1434,10 @@ class TwitchChatDownloader(BaseChatDownloader):
         author_badge_metadata = info.pop('author_badge_metadata', [])
         author_badges = info.pop('author_badges', [])
 
-        info['author_badges'] = TwitchChatDownloader.parse_irc_badges(
+        info['author_badges'] = TwitchChatDownloader._parse_irc_badges(
             author_badges, info.get('channel_id'))
 
-        badge_metadata = TwitchChatDownloader.parse_irc_badges(
+        badge_metadata = TwitchChatDownloader._parse_irc_badges(
             author_badge_metadata, info.get('channel_id'))
 
         subscriber_badge = next(
@@ -1319,10 +1452,10 @@ class TwitchChatDownloader(BaseChatDownloader):
         if author_display_name:
             info['author_name'] = author_display_name.lower()
 
-        in_reply_to = BaseChatDownloader.move_to_dict(info, 'in_reply_to')
+        in_reply_to = BaseChatDownloader._move_to_dict(info, 'in_reply_to')
 
-        BaseChatDownloader.move_to_dict(in_reply_to, 'author')
-        BaseChatDownloader.move_to_dict(info, 'author')
+        BaseChatDownloader._move_to_dict(in_reply_to, 'author')
+        BaseChatDownloader._move_to_dict(info, 'author')
 
         original_action_type = match.group(2)
 
@@ -1380,13 +1513,6 @@ class TwitchChatDownloader(BaseChatDownloader):
         messages_groups_to_add = params.get('message_groups') or []
         messages_types_to_add = params.get('message_types') or []
 
-        def debug_log(*items):
-            log(
-                'debug',
-                items,
-                params.get('pause_on_debug')
-            )
-
         def create_connection():
             for attempt_number in attempts(max_attempts):
                 try:
@@ -1394,7 +1520,7 @@ class TwitchChatDownloader(BaseChatDownloader):
                     irc.set_timeout(message_receive_timeout)
                     irc.join_channel(stream_id)
                     return irc
-                except socket.gaierror as e:
+                except (socket.gaierror, ConnectionRefusedError) as e:
                     self.retry(attempt_number, max_attempts, e, retry_timeout)
 
         twitch_chat_irc = create_connection()
@@ -1408,12 +1534,8 @@ class TwitchChatDownloader(BaseChatDownloader):
 
         message_count = 0
 
-        timeout = Timeout(params.get('timeout'))
-        inactivity_timeout = Timeout(params.get(
-            'inactivity_timeout'), Timeout.INACTIVITY)
         try:
             while True:
-                timeout.check_for_timeout()
 
                 try:
                     new_info = twitch_chat_irc.recv(buffer_size)
@@ -1465,15 +1587,16 @@ class TwitchChatDownloader(BaseChatDownloader):
                             missing_keys = data.keys() - TwitchChatDownloader._KNOWN_IRC_KEYS
 
                             if missing_keys:
-                                debug_log(
-                                    'Missing keys found: {}'.format(
-                                        missing_keys),
-                                    'Original data: {}'.format(match.groups()),
-                                    'Parsed data: {}'.format(data)
-                                )
+                                self._debug_log(params,
+                                                'Missing keys found: {}'.format(
+                                                    missing_keys),
+                                                'Original data: {}'.format(
+                                                    match.groups()),
+                                                'Parsed data: {}'.format(data)
+                                                )
                             # check whether to skip this message or not, based on its type
 
-                            to_add = self.must_add_item(
+                            to_add = self._must_add_item(
                                 data,
                                 self._MESSAGE_GROUPS,
                                 messages_groups_to_add,
@@ -1483,7 +1606,6 @@ class TwitchChatDownloader(BaseChatDownloader):
                             if not to_add:
                                 continue
 
-                            inactivity_timeout.reset()
                             message_count += 1
                             yield data
 
@@ -1509,7 +1631,7 @@ class TwitchChatDownloader(BaseChatDownloader):
                         last_ping_time = current_time
 
                 except socket.timeout:
-                    inactivity_timeout.check_for_timeout()
+                    pass  # Allows for keyboard interrupts
 
                 except ConnectionError:
                     # Close old connection
@@ -1521,6 +1643,9 @@ class TwitchChatDownloader(BaseChatDownloader):
         finally:
             twitch_chat_irc.close_connection()
 
+    def _get_chat_by_stream_id(self, match, params):
+        return self.get_chat_by_stream_id(match.group('id'), params)
+
     def get_chat_by_stream_id(self, stream_id, params):
 
         max_attempts = params.get('max_attempts')
@@ -1535,9 +1660,12 @@ class TwitchChatDownloader(BaseChatDownloader):
             try:
                 stream_info = self._download_gql(query)[0]['data']['user']
                 break
-            except (UnexpectedHTML, RequestException) as e:
+            except (JSONDecodeError, RequestException) as e:
                 self.retry(attempt_number, max_attempts, e, retry_timeout)
 
+        if not stream_info:
+            raise UserNotFound('Unable to find user: "{}"'.format(stream_id))
+
         is_live = multi_get(stream_info, 'stream', 'type') == 'live'
         title = multi_get(stream_info, 'lastBroadcast',
                           'title') if is_live else None
@@ -1552,27 +1680,27 @@ class TwitchChatDownloader(BaseChatDownloader):
             is_live=is_live
         )
 
-    def get_chat(self,
-                 **kwargs
-                 ):
+    # def get_chat(self,
+    #              **kwargs
+    #              ):
 
-        # get video id
-        url = kwargs.get('url')
+    #     # get video id
+    #     url = kwargs.get('url')
 
-        for regex, function_name in self._REGEX_FUNCTION_MAP:
-            match = re.search(regex, url)
-            if match:
-                return getattr(self, function_name)(match.group('id'), kwargs)
+    #     for regex, function_name in self._REGEX_FUNCTION_MAP:
+    #         match = re.search(regex, url)
+    #         if match:
+    #             return getattr(self, function_name)(match.group('id'), kwargs)
 
         # if(match):
         #     match.group('id')
         #     return self.get_chat_by_video_id(match.group('id'), params)
 
-            # if(match.group('id')):  # normal youtube video
-            #     return
+        # if(match.group('id')):  # normal youtube video
+        #     return
 
-            # else:  # TODO add profile, etc.
-            #     pass
+        # else:  # TODO add profile, etc.
+        #     pass
 
     # def get_chat_messages(self, url):
     #     pass
